"""
Supports the discrimination script
"""

from ase.io import read, write, iread
import numpy as np
from tqdm import tqdm
from ase import Atoms, Atom
import math
from typing import Tuple, Literal, List, Union
from ase.constraints import FixAtoms
from ase.neighborlist import NeighborList, natural_cutoffs
import numpy as np
from ase.geometry.analysis import Analysis

LATERAL_SPACING = (
    14.5  # spacing between between each nanoparticles repeated image. Note, though that
)
# this is a little lower than the spacing, due to the ceil function being used
Z_SPACING = 13.5  # spacing in z direction
ADSORPTION_HEIGHT = 2.2  # adsorption height
INTERFACE_SCALING = 1.3  # scaling factor for interfacial NP atoms' covalent radii
NANOPARTICLE_ELEMENT = "Ag"  # what kind of atom is the NP made of?
SCALING_FACTOR = 1.0
SKIN = 0.3  # ASE's default value=0.3A


def create_unit_support(support: Literal["graphene", "au", "mgo"]) -> Atoms:
    """Creates a unit cell of the support: graphene, mgo, or gold"""
    support = support.lower()

    graphene = Atoms(
        "C4",
        pbc=np.array([True, True, True]),
        cell=np.array(
            [
                [4.2687822100000004, 0.0000000000000000, 0.0000000000000000],
                [0.0000000000000000, 2.4645825600000002, 0.0000000000000000],
                [0.0000000000000000, 0.0000000000000000, 9.5824915799999992],
            ]
        ),
        positions=np.array(
            [
                [1.4229274000000001, 0.0000000000000000, 0.0000000000000000],
                [2.8458548100000001, 0.0000000000000000, 0.0000000000000000],
                [0.7114637000000000, 1.2322912800000001, 0.0000000000000000],
                [3.5573185100000000, 1.2322912800000001, 0.0000000000000000],
            ]
        ),
    )

    gold = Atoms(
        "Au4",
        pbc=np.array([True, True, True]),
        cell=np.array(
            [
                [2.8849956672411143, 0.0000000000000000, 0.0000000000000000],
                [0.0000000000000000, 4.9969590752776831, 0.0000000000000000],
                [0.0000000000000000, 0.0000000000000000, 6.3555890982936738],
            ]
        ),
        positions=np.array(
            [
                [0.0000000000000000, 1.6656530250925610, 2.0000000000000000],
                [1.4424978336205572, 4.1641325627314023, 2.0000000000000000],
                [0.0000000000000000, 0.0000000000000000, 4.3555890982936738],
                [1.4424978336205572, 2.4984795376388416, 4.3555890982936738],
            ]
        ),
    )

    mg = Atoms(
        "Mg4",
        pbc=graphene.pbc,
        cell=np.array(
            [
                [4.2602238819630642, 0.0000000000000000, 0.0000000000000000],
                [0.0000000000000003, 4.2602238819630642, 0.0000000000000000],
                [0.0000000000000000, 0.0000000000000000, 24.9107835868707248],
            ]
        ),
        positions=np.array(
            [
                [0.0000000000000021, 0.0000000000000009, 17.7806716458891927],
                [2.1301119409815343, 2.1301119409815330, 17.7806716458891927],
                [0.0000000000000025, 2.1301119409815330, 19.9107835868707248],
                [2.1301119409815348, 0.0000000000000011, 19.9107835868707248],
            ]
        ),
    )

    o = Atoms(
        "O4",
        pbc=graphene.pbc,
        cell=mg.cell,
        positions=np.array(
            [
                [0.0000000000000023, 2.1301119409815330, 17.7806716458891927],
                [2.1301119409815343, 0.0000000000000009, 17.7806716458891927],
                [0.0000000000000025, 0.0000000000000011, 19.9107835868707248],
                [2.1301119409815348, 2.1301119409815334, 19.9107835868707248],
            ]
        ),
    )

    mg.extend(o)

    requested = {"graphene": graphene, "au": gold, "mgo": mg}
    unit_support = requested[support]

    return unit_support


def setup_neighborlist(
    atoms: Atoms,
    scaling_factor: float = SCALING_FACTOR,
    radial_cutoff: float = None,
    skin: float = SKIN,
):
    """
    Setup neighborlist object,
    given the Atoms object and the scaling factor for covalent radii

    if radial_cutoff is given, then we will consider only neighbors within
    the lower of that cutoff and the bond length
    """
    if not radial_cutoff:
        cutoffs = natural_cutoffs(atoms, mult=scaling_factor)
    else:
        cutoffs = np.minimum(radial_cutoff / 2, np.array(natural_cutoffs(atoms)))

    neighbors = NeighborList(
        cutoffs=cutoffs, self_interaction=False, bothways=True, skin=skin
    )
    neighbors.update(atoms)
    return neighbors


def setup_analyzer(atoms: Atoms, neighborlist: NeighborList = None) -> Analysis:
    """
    Setup the analyzer object, given an optional NeighborList object
    If not given a NeighborList object, use defaults
    """
    analyzer = Analysis(atoms, nl=neighborlist)
    return analyzer


def divider(
    atoms: Atoms, element: str = NANOPARTICLE_ELEMENT
) -> Union[None, Tuple[Atoms]]:
    """
    Divide the structure into NP and surface
    Preserve the constraints of the support; assume the NP has none

    Inputs:
                Supported_NP
                element         (what element is the NP?)
                                default = 'Ag'
    Returns:    NP
                Bare_Surface
    """
    if not isinstance(atoms, Atoms):
        return None

    element = element.capitalize()

    constrained_indices = (
        atoms.constraints[0].get_indices() if atoms.constraints else list()
    )
    silver_indices = [
        index for index, atom in enumerate(atoms) if atom.symbol == element
    ]
    support_indices = [
        index for index, atom in enumerate(atoms) if atom.symbol != element
    ]

    silvers, support = atoms[silver_indices], atoms[support_indices]
    silvers.set_cell(atoms.get_cell())
    support.set_cell(atoms.get_cell())
    silvers.pbc, support.pbc = True, True

    old_to_new_support_indices = dict()
    for new_index, old_index in enumerate(support_indices):
        old_to_new_support_indices[old_index] = new_index

    new_constrained_support_indices = [
        old_to_new_support_indices[old_i]
        for old_i in constrained_indices
        if old_i in old_to_new_support_indices
    ]

    if new_constrained_support_indices:
        support.set_constraint(FixAtoms(indices=new_constrained_support_indices))

    return silvers, support


def centralize(atoms: Atoms, element: str = NANOPARTICLE_ELEMENT) -> Atoms:
    """
    Centralize the NP to make viewing easier
    Inputs:     Supported_NP
                element the NP is made of
    Returns:    Supported_NP (with NP centralized)
    """
    silvers, support = divider(atoms, element=element)

    center_x = support.cell.lengths()[0] / 2
    center_y = support.cell.lengths()[1] / 2

    silvers_com_x = silvers.get_center_of_mass()[0]
    silvers_com_y = silvers.get_center_of_mass()[1]

    x_move = center_x - silvers_com_x
    y_move = center_y - silvers_com_y
    silvers.translate(displacement=(x_move, y_move, 0))

    support.extend(silvers)
    support.pbc = atoms.pbc
    support.set_constraint(atoms.constraints)

    support.pbc = True

    return support


def scaler(
    image: Atoms,
    adsorption_height: float = ADSORPTION_HEIGHT,
    z_spacing: float = Z_SPACING,
    lateral_spacing: float = LATERAL_SPACING,
    unit_support: Union[Atoms, None] = None,
) -> Atoms:
    """
    Return an NP supported upon a support of a good size
    Inputs:
        Image:              NP or Supported NP
        element:            What element the NP is made of. default = 'Ag'
        adsorption_height:  Desired adsorption height (Optional)
        z_spacing:          Desired Z-spacing of periodic images (Optional)
        lateral_spacing:    Desired X- and Y-spacing of periodic NP images (Optional)
        unit_support:       Atoms object of the unit support.
    Returns:
                            Supported_NP
    """
    if not unit_support:
        unit_support = create_unit_support(support="mgo")
    unit_cell = unit_support.cell
    unit_cell_x = unit_cell[0, 0]
    unit_cell_y = unit_cell[1, 1]
    unit_cell_z = unit_cell[2, 2]
    unit_cell_max_z = max(unit_support.positions[:, 2])

    silvers, support = divider(image, element=image.get_chemical_symbols()[0])
    silvers = Atoms(silvers)
    silvers.center(vacuum=10)

    min_x = min(silvers.positions[:, 0])
    min_y = min(silvers.positions[:, 1])
    max_x = max(silvers.positions[:, 0])
    max_y = max(silvers.positions[:, 1])
    min_z = min(silvers.positions[:, 2])

    x_diameter = max_x - min_x
    y_diameter = max_y - min_y

    required_x = x_diameter + lateral_spacing
    required_y = y_diameter + lateral_spacing

    ratio_x = math.ceil(required_x / unit_cell_x)
    ratio_y = math.ceil(required_y / unit_cell_y)

    adsorption_height = adsorption_height + unit_cell_max_z
    new_support = unit_support * (ratio_x, ratio_y, 1)
    silvers_displacement = adsorption_height - min_z
    silvers.translate(displacement=(0, 0, silvers_displacement))
    # I think the logic here is tedious, but never mind
    new_support.cell[2, 2] = (
        10  # deliberately set too low so that the logic within the if statement (below) will play out fine
    )
    new_support.extend(silvers)

    cell_bottom = 0
    min_z = min(new_support.positions[:, 2])
    new_cell_displacement = cell_bottom - min_z
    new_support.translate(displacement=(0, 0, new_cell_displacement))

    max_height = max(new_support.positions[:, 2])
    cell_top = new_support.cell[2, 2]
    distance = cell_top - max_height

    if distance < z_spacing:
        new_support.cell[2, 2] += lateral_spacing - cell_top + max_height

    new_support = centralize(new_support)
    new_support.info.update(image.info)

    return new_support



"""
In memory of Korosensei

This is the atomistic method

Use:                    Discrimination of kinds of atoms in supported NPs
                        The interface absolutely MUST be flat or nearly so

Functions:              get_interface
                        get_np_surface_by_CN
                        get_perimeter
                        discriminate


To remove complications regarding alphashape and its very specific set of dependencies
(and on how slow optimizing alpha is for large nanoparticles)
I have eliminated it, commenting it out entirely

Just to note it here, in case a future version has me putting back alphashape, here's the process:
    pip uninstall alphashape
    pip install git+https://github.com/bellockk/alphashape.git
    pip install trimesh==4.2.0 numpy==1.26.4

"""

import pandas as pd
from argparse import ArgumentParser
from copy import deepcopy
from ase.build import bulk
from tqdm import tqdm
import statistics
from ase.io import read, write
import numpy as np
from ase import Atoms, Atom
import math
from sys import argv, exit

# import alphashape
# from shapely.geometry import Polygon, MultiPolygon
from ase.geometry.analysis import Analysis
from typing import Tuple, Literal, List, Union
from ase.neighborlist import natural_cutoffs, NeighborList
import warnings
from os import system, environ
from benchmark.atomistic_utils import (
    divider,
    setup_neighborlist,
    setup_analyzer,
    NANOPARTICLE_ELEMENT,  # SKIN
)
from joblib import Parallel, delayed
from ascii_colors import ASCIIColors


INTERFACE_SCALING = 1.3  # 1.3 scaling factor for interfacial NP atoms' covalent radii
CN_SCALING = 1.05  # 1.2 scaling factor for discriminating bulk from NP surface atoms
FCC_AND_HCP_COORD_CUTOFF = 12  # any atom with coord < COORD_CUTOFF is a surface atom
BCC_COORD_CUTOFF = 8  # any atom with coord < COORD_CUTOFF is a surface atom
INTERFACIAL_FACET = (1, 1, 1)  # facet of NP that faces the support;
# also that which is on the outer surface of the NP
# which means this is a misnomer
DO_ALPHA_SHAPE = False  # don't use alpha shape to get perimeter atoms
PROCESSES = -1  # How many processes to run in parallel.
NL_SKIN = 0.05  # Ang; nl skin for distinguishing surface NP atoms;
# set low because we normally use this on perfect crystals,
# not structures from MD; this value seems to work well


def get_interface(
    atoms: Atoms,
    nl: NeighborList = None,
    analyzer: Analysis = None,
    support_elements: Union[str, List[str]] = None,
    np_element: str = NANOPARTICLE_ELEMENT,
    scaling_factor: float = INTERFACE_SCALING,
) -> Union[List[None], Tuple[Atoms, Atoms, List[int], List[int]]]:
    """
    Get indices of interfacial atoms between NP and Support

    Requires:
        atoms (Atoms):                      atoms object of NP + Support
        nl (NeighborList):                  neighborlist object of whole system (i.e. NP + Support).
                                            Optional but strongly recommended to provide analyzer or nl
        analyzer (Analysis):                Analysis object  of whole system (i.e. NP + Support).
                                            Optional but strongly recommended to provide analyzer or nl
        support_elements (Union[str, List[str]]):
                                            what element(s) the support is made of. If not supplied, we
                                            say it is whatever np_element is not
        np_element (str):                   what element the NP is made of. default = 'Ag'
        scaling_factor (float):             scaling factor for covalent radii. Optional
                                            Advised to be > 1.0 to fully capture the interface
    Returns:
        np_interfacial_indices (List[int]):         indices of interfacial (+ perimeter) NP atoms
        support_interfacial_indices (List[int]):    indices of interfacial support atoms

    If no NP-Support bonds found,
        Returns:    None, None
    """
    elements = np.unique(atoms.get_chemical_symbols())
    np_element = np_element.capitalize()
    if not support_elements:
        support_elements = [i for i in elements if i != np_element]
    if isinstance(support_elements, str):
        support_elements = [support_elements]

    if not analyzer:
        if not nl:  # reuse neighborlist if given
            natural_cutoff = (
                np.max([natural_cutoffs(Atoms(element)) for element in elements]) * 2
            )  # no need to search beyond this distance
            nl = setup_neighborlist(atoms, scaling_factor, radial_cutoff=natural_cutoff)
            nl.update(atoms)

        analyzer = setup_analyzer(
            atoms, neighborlist=nl
        )  # reuse Analysis object if given

    NP_Support_bonds = {
        e.capitalize(): analyzer.get_bonds(e.capitalize(), np_element, unique=True)[0]
        for e in support_elements
    }
    support_interfacial_indices, np_interfacial_indices = [], []
    for support_e, bonds in NP_Support_bonds.items():
        if bonds:
            np_interfacial_indice, support_interface_indice = zip(
                *[(j, i) for (i, j) in bonds]
            )
            np_interfacial_indice = list(np_interfacial_indice)
            support_interface_indice = list(support_interface_indice)
        else:
            warnings.warn(
                f"""There are no {support_e}-{np_element} bonds in your system!
            This may be a very small cluster""",
                category=RuntimeWarning,
            )
            np_interfacial_indice = list()
            support_interface_indice = list()

        support_interfacial_indices.extend(support_interface_indice)
        np_interfacial_indices.extend(np_interfacial_indice)

    if not (support_interfacial_indices and np_interfacial_indices):
        print(f"No NP-Support bonds found! Failed!")
        return None, None

    support_interfacial_indices = list(np.unique(support_interfacial_indices))
    np_interfacial_indices = list(np.unique(np_interfacial_indices))

    # CAUTION!! np_interfacial_indices includes perimeter atoms.
    # To separate out the perimeter, use get_perimeter on results from this function
    return np_interfacial_indices, support_interfacial_indices


def get_perimeter(
    atoms: Atoms,
    interfacial_results: [List[int], List[int]],
    do_alphashape: bool = DO_ALPHA_SHAPE,
) -> Union[List[None], Tuple[Atoms, Atoms, List[int], List[int], float]]:
    f"""
    Get indices of perimeter atoms of the NP
    By:
        1. Getting the interface
        2. Convex/concave hull of interfacial atoms to trace the perimeter (misses concave regions);
        from experience, the bigger the NP, the more it'll undercount the perimeter, especially the convex
        hull.
        3. To correct (2), also gets CNmax (the most-coordinated atom amongst those flagged by (2)
        as perimeter) minus 1, and defines that any atom of coordination less than the max of CNmax and
        CNmedian (median coordination of all atoms flagged by (2) as perimeter).
        i.e.:
            threshold_CN = max(CNmedian, CNmax - 1)

        Note: simply setting threshold_CN = CNmax would be good but a little risky in rare cases

    Requires:
        atoms (Atoms):                      atoms object of NP + Support
        interfacial_results: Tuple[List[int], List[int]]. Required; results of the get_interface function,
        do_alphashape (bool):               whether or not to get perimeter atoms by alphashape rather than only by CN
                                            default = False
                                            It's very expensive to optimize alpha for large NPs

    Returns:
        perimeter_indices (List[int])      indices of perimeter atoms

    If get_interface function failed,
        Returns:    None

    Warning:
        This threshold_CN method will only work when the atomic density
        at the interface is homogeneous
    """
    np_interfacial_indices, support_interfacial_indices = interfacial_results
    np_interface = atoms[np_interfacial_indices]  # NPinterface (i.e. no support)

    # get perimeter atoms
    perimeter_indices = list()

    # criterion 1: concave hull vertices
    # August 1st, 2025: hardcoded to never run
    if False:  # do_alphashape:
        positions_2D = np_interface.positions[:, :2]
        optimum_alpha = alphashape.optimizealpha(positions_2D)
        print(f"Optimal alpha = {optimum_alpha}")
        alpha_shape = alphashape.alphashape(
            positions_2D, optimum_alpha * 0.88
        )  # scale down optimum_alpha because it'll tend to overcount
        perim_set = {tuple(pt) for pt in alpha_shape.exterior.coords[:-1]}
        perimeter_indices = [
            i for (i, p) in enumerate(positions_2D) if tuple(p) in perim_set
        ]

    # criterion 2: coordination threshold
    ##we have to create a new nl because we are now using a subset of the provided Atoms object
    neighbors = NeighborList(
        cutoffs=natural_cutoffs(np_interface, mult=INTERFACE_SCALING),
        self_interaction=False,
        bothways=True,
    )
    neighbors.update(np_interface)
    CNs = [
        (atom_index, len(neighbors.get_neighbors(atom_index)[0]))
        for (atom_index, _) in enumerate(np_interface)
    ]  # get number of neighbors for each interfacial NP atom
    # remember: only interface-interface bonds are considered
    neighbor_counts = [num_neighbors for (_, num_neighbors) in CNs]
    medianCN, maxCN = np.median(neighbor_counts), max(neighbor_counts)
    thresholdCN = max(medianCN, maxCN)  # - 1)
    missed_perimeter_indices = [
        atom_index
        for (atom_index, num_neighbors) in CNs
        if (num_neighbors < thresholdCN and atom_index not in perimeter_indices)
    ]
    perimeter_indices.extend(missed_perimeter_indices)

    # convert the perimeter indices to the indices in the Atoms object
    perimeter_indices = [np_interfacial_indices[i] for i in perimeter_indices]

    return perimeter_indices


def get_np_surface_by_CN(
    atoms: Atoms,
    nl: NeighborList = None,
    support_elements: Union[str, List[str]] = None,
    np_element: str = NANOPARTICLE_ELEMENT,
    scaling_factor: float = CN_SCALING,
    coord_cutoff: int = None,
) -> Tuple[List[int], List[int]]:
    """
    Separate NP surface's from bulk based on CN (coordination numbers)

    Requires:
        atoms (Atoms):                      atoms object of NP + Support
        nl (NeighborList):                  neighborlist object of whole system (i.e. NP + Support).
                                            Optional but strongly recommended to provide analyzer or nl
        support_elements (Union[str, List[str]]):
                                            what element(s) the support is made of.
        np_element (str):                   what element the NP is made of. Default = 'Ag'
        scaling_factor (float):             scaling factor for covalent radii.
                                            optional. default = 1.05
        coord_cutoff (int):                 coordination cutoff for discriminating surface from bulk atoms
                                            optional. default = 12 for FCC and 8 for BCC
    Returns
        surface_plus_interface_indices (List[int]):
                                            indices of nanoparticle surface atoms.
                                            Note: This INCLUDES THE INTERFACE + PERIMETER ATOMS
    """
    if not support_elements:
        support_elements = [i for i in elements if i != np_element]
    if isinstance(support_elements, str):
        support_elements = [support_elements]
    if not nl:
        # create neighborlist object
        cutoff_kwargs = {element.capitalize(): 0 for element in support_elements}
        scaling_factor = 1.0
        cutoffs = natural_cutoffs(
            atoms, mult=scaling_factor, **cutoff_kwargs
        )  # set non NP atoms to zero cutoffs
        nl = NeighborList(
            cutoffs=cutoffs, self_interaction=False, bothways=True, skin=NL_SKIN
        )
        nl.update(atoms)

    # apply criterion
    if not coord_cutoff:
        lattice = bulk(np_element).cell.get_bravais_lattice().__class__.__name__
        if lattice in ["FCC", "HCP", "HEX"]:  # should confirm later, but it seems ASE
            # has only hex not hcp (they mean the same, I suppose)
            coord_cutoff = FCC_AND_HCP_COORD_CUTOFF
        elif lattice == "BCC":
            coord_cutoff = BCC_COORD_CUTOFF
        else:
            warnings.warn(
                f"""NP element's crystal system is neither fcc, hcp, nor bcc!
            Setting coordination cutoff to that of fcc/hcp ({FCC_AND_HCP_COORD_CUTOFF})!
            """,
                category=UserWarning,
            )

    CN = [
        (index, len(nl.get_neighbors(index)[0]))
        for index, i in enumerate(atoms)
        if i.symbol == np_element
    ]
    bulk_indices = [i for i, j in CN if j == coord_cutoff]
    surface_plus_interface_indices = [i for i, j in CN if i not in bulk_indices]

    return surface_plus_interface_indices


def discriminate(
    atoms: Atoms,
    nl: NeighborList,
    analyzer: Analysis,
    support_elements: Union[str, List[str]],
    np_element: str = NANOPARTICLE_ELEMENT,
    interface_scaling: float = INTERFACE_SCALING,
    surface_scaling: float = CN_SCALING,
    coord_cutoff: int = None,
    do_alphashape: bool = DO_ALPHA_SHAPE,
) -> Tuple[
    List[int],
    List[int],
    List[int],
    List[int],
    List[int],
]:
    """Main function for separating a supported NP into NP surface, NP Bulk, NP interface (non-perimeter),
    NP perimeter, and Support

    Requires:
        atoms (Atoms):                      atoms object of NP + Support
        nl (NeighborList):                  neighborlist object of whole system (i.e. NP + Support).
                                            Optional but strongly recommended to provide analyzer or nl
        analyzer (Analysis):                Analysis object  of whole system (i.e. NP + Support).
                                            Optional but strongly recommended to provide analyzer or nl
        support_elements (Union[str, List[str]]):
                                            what element(s) the support is made of.
        np_element (str):                   what element the NP is made of. default = 'Ag'
        interface_scaling (float):          scaling factor for covalent radii to discriminate interface from rest of system.
                                            default = 1.3
        surface_scaling (float):             scaling factor for covalent radii. Default = 1.05
        coord_cutoff (int):                 coordination cutoff for discriminating surface from bulk atoms
                                            optional. default = 12 for FCC and 8 for BCC
        do_alphashape (bool):               whether or not to get perimeter atoms by alphashape rather than only by CN
                                            default = False
                                            !!!It's very expensive to optimize alpha for large NPs!!!
                                            So, I hope you know what you are doing!

    Returns:
        np_surface (List[int]):             indices of NP surface atoms (excluding interface)
        np_bulk (List[int]):                indices of bulk NP atoms
        np_interface (List[int]):           indices of NP interfacial atoms (excluding perimeter)
        np_perimeter (List[int]):           indices of NP perimeter atoms
        substrate_interface (List[int]):     indices of support interfacial atoms
        substrate (List[int]):              indices of support (including interfacial support) atoms

    Returns [0,0,0,0,0,0] if get_interface didnt get any NP-support bonds
    """
    do_alphashape = False  # hardcoded to False (August 2025)

    elements = np.unique(atoms.get_chemical_symbols())
    np_element = np_element.capitalize()
    if not support_elements:
        support_elements = [i for i in elements if i != np_element]
    if isinstance(support_elements, str):
        support_elements = [support_elements.capitalize()]
    else:
        support_elements = [i.capitalize() for i in support_elements]
    np_and_support_elements = support_elements + [np_element]
    check_elements = [i in elements for i in np_and_support_elements]
    if not np.all(check_elements):
        raise ValueError(
            f"Some of the given elements ({np_and_support_elements}) not in Atoms object!"
        )

    if not analyzer:
        if not nl:  # reuse neighborlist if given
            natural_cutoff = (
                np.max([natural_cutoffs(Atoms(element)) for element in elements]) * 2
            )  # no need to search beyond this distance
            nl = setup_neighborlist(
                atoms, scaling_factor=interface_scaling, radial_cutoff=natural_cutoff
            )  # , skin = NL_SKIN) adding this ruins things!
            nl.update(atoms)

        analyzer = setup_analyzer(
            atoms, neighborlist=nl
        )  # reuse Analysis object if given

    nps = [index for index, _ in enumerate(atoms) if atoms[index].symbol == np_element]
    substrate = [
        index
        for index, _ in enumerate(atoms)
        if atoms[index].symbol in support_elements
    ]

    try:
        np_interface_plus_perimeter, substrate_interface = get_interface(
            atoms,
            nl=nl,
            analyzer=analyzer,
            np_element=np_element,
            support_elements=support_elements,
            scaling_factor=interface_scaling,
        )

        np_perimeter = get_perimeter(
            atoms,
            interfacial_results=(np_interface_plus_perimeter, substrate_interface),
            do_alphashape=do_alphashape,
        )

        np_interface = [i for i in np_interface_plus_perimeter if i not in np_perimeter]
        nl = None  # unfortunately we have to reset in order to set skin to zero for the surface discriminaion
        np_surface_plus_interface = get_np_surface_by_CN(
            atoms,
            nl=nl,
            np_element=np_element,
            support_elements=support_elements,
            scaling_factor=surface_scaling,
            coord_cutoff=coord_cutoff,
        )

        np_surface = [
            i for i in np_surface_plus_interface if i not in np_interface_plus_perimeter
        ]
        np_bulk = [i for i in nps if i not in np_surface_plus_interface]

    except TypeError:  # get_interface returned None, None
        ASCIIColors.red("get_interface probably failed")
        (
            np_surface,
            np_bulk,
            np_interface,
            np_perimeter,
            substrate_interface,
            substrate,
        ) = [0] * 6
    except Exception as e:
        ASCIIColors.red(f"Unexpected error occurred:\t{e}")
        exit(1)

    return (
        np_surface,
        np_bulk,
        np_interface,
        np_perimeter,
        substrate_interface,
        substrate,
    )


if __name__ == "__main__":
    parser = ArgumentParser(
        description="""Splits up a supported NP into surface, substrate, bulk, perimeter, and interface
    Also calculates interfacial radius by any/both of two methods if requested"""
    )

    parser.add_argument(
        "--traj",
        "-t",
        type=str,
        required=True,
        help="Traj file on which to run the script. Required",
    )
    parser.add_argument(
        "--traj_output",
        "-to",
        type=str,
        required=True,
        help="Traj file to create in which atoms have been discriminated. Required",
    )
    parser.add_argument(
        "--processes",
        "-p",
        type=int,
        default=PROCESSES,
        help=f"How many processes to run in parallel. default = {PROCESSES}",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default="output_atomistic.csv",
        help="Output filename.",
    )
    parser.add_argument(
        "--np_element",
        "-ne",
        type=str,
        default=None,  # NANOPARTICLE_ELEMENT,
        help=f"Element of which the NP is composed. by default we'll read atoms.info['np_element']",
    )

    parser.add_argument(
        "--support_elements",
        "-se",
        type=str,
        nargs="+",
        default=None,
        help="""List of elements of which the substrate is composed.
    In the format: a b c d. Defaults to elements absent from the NP""",
    )

    parser.add_argument(
        "--interface_scaling",
        "-is",
        type=float,
        default=INTERFACE_SCALING,
        help=f"""scaling factor for interfacial atoms' covalent radii.
            Default = {INTERFACE_SCALING}""",
    )
    parser.add_argument(
        "--surface_scaling",
        "-ss",
        type=float,
        default=CN_SCALING,
        help=f"""scaling factor for discriminating NP surface from bulk.
            Default = {CN_SCALING}""",
    )

    do_alphashape = False

    args = parser.parse_args()
    output_file = args.output
    output_traj = args.traj_output
    atoms = read(args.traj, ":")
    atoms = atoms if isinstance(atoms, list) else [atoms]
    processes = environ.get("SLURM_NTASKS_PER_NODE", args.processes)
    processes = int(processes)
    np_element_list = [
        image.info["np_element"] if not args.np_element else args.np_element
        for image in atoms
    ]

    with Parallel(n_jobs=processes) as parallel:
        results = parallel(
            delayed(discriminate)(
                image,
                np_element=np_element_list[index],
                support_elements=args.support_elements,
                interface_scaling=args.interface_scaling,
                surface_scaling=args.surface_scaling,
                do_alphashape=do_alphashape,  # args.do_alphashape,
                nl=None,
                analyzer=None,
            )
            for index, image in enumerate(
                tqdm(atoms, desc="Discriminating..", total=len(atoms))
            )
        )

    np_surface, np_bulk, np_interface, np_perimeter, substrate_interface, substrate = (
        zip(*results)
    )
    np_total = [
        (j + np_bulk[i] + np_perimeter[i] + np_interface[i])
        for i, j in enumerate(np_surface)
    ]

    np_interface_plus_peri = [j + np_perimeter[i] for i, j in enumerate(np_interface)]
    interfacial_results_list = [
        (j, substrate_interface[i]) for i, j in enumerate(np_interface_plus_peri)
    ]

    # return Traj but with atoms' identities switched
    for index, image in enumerate(
        tqdm(atoms, total=len(atoms), desc="Changing identities..")
    ):
        np_surf, np_int, np_peri = (
            np_surface[index],
            np_interface[index],
            np_perimeter[index],
        )

        # hardcoded; should include in args sometime but this is sufficient for tests
        for i in np_int:
            image[i].symbol = "Cu"
        for i in np_peri:
            image[i].symbol = "Mo"
        for i in np_surf:
            image[i].symbol = "Pd"

    print("Saving results")
    write(output_traj, atoms)

    ##time to write results
    # write the output for the atomistic_discriminator
    np_perimeter_len = [len(i) for i in np_perimeter]
    np_interface_len = [len(i) for i in np_interface]
    np_surface_len = [len(i) for i in np_surface]
    np_total_len = [len(i) for i in np_total]

    df = pd.DataFrame(
        {
            "Perimeter": np_perimeter_len,
            "Interface": np_interface_len,
            "Surface": np_surface_len,
            "Total": np_total_len,
        }
    )
    df.to_csv(output_file, index=False)



"""
Automatically test all functions
"""

from pathlib import Path
import sys
from ascii_colors import ASCIIColors
import pytest


def main() -> None:
    """
    Run every test, then exit with pytest’s status code.
    """
    repo_root = Path(__file__).resolve().parents[2]
    tests_dir = repo_root / "tests"
    exit_code = pytest.main(["-s", str(tests_dir)])
    text = "All tests passed!" if exit_code == 0 else "Some tests failed!"
    color = ASCIIColors.color_green if exit_code == 0 else ASCIIColors.color_red
    ASCIIColors.print(
        text.upper(),
        color=color,
        style=ASCIIColors.style_bold,
        background=ASCIIColors.color_black,
        end="\n\n",
        flush=True,
        file=sys.stdout,
    )
    sys.exit(exit_code)  # sys.exit(pytest.main(["-s", str(tests_dir)]))


if __name__ == "__main__":
    main()



"""
Plot a parity line for the data from two CSV files.
Here, they'll be the outputs of the atomcounter vs the atomistic model

The two input files must share the same column headers
(e.g., Perimeter, Interface, Surface, Total).
"""

import sys
from argparse import ArgumentParser
from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt


def read_args():
    parser = ArgumentParser()
    parser.add_argument("csv_a", help="first CSV file. Atomistic output")
    parser.add_argument("csv_b", help="second CSV file. AtomCounter output")
    parser.add_argument(
        "--show",
        action="store_true",
        help="show the plots on screen (in addition to saving)",
    )
    parser.add_argument(
        "--output_dir", type=str, help="directory where parity plots should be saved"
    )

    return parser.parse_args()


def main():
    args = read_args()
    df_a = pd.read_csv(args.csv_a)
    df_b = pd.read_csv(args.csv_b)

    df_a.columns = df_a.columns.str.strip()
    df_b.columns = df_b.columns.str.strip()

    common_cols = df_a.columns.intersection(df_b.columns)
    if common_cols.empty:
        sys.exit("Error: the two files share no common column headers.")

    out_dir = args.output_dir

    ##create parity plots
    for col in common_cols:
        x = df_a[col]
        y = df_b[col]

        lo = min(x.min(), y.min())
        hi = max(x.max(), y.max())
        pad = 0.05 * (hi - lo)
        lo, hi = lo - pad, hi + pad

        plt.figure(figsize=(4, 4))
        plt.scatter(x, y, c="tab:blue")
        plt.plot([lo, hi], [lo, hi], "k--", lw=1)
        plt.xlabel(f"{col} – file A")
        plt.ylabel(f"{col} – file B")
        plt.title(f"Parity plot of results: {col}")
        plt.xlim(lo, hi)
        plt.ylim(lo, hi)
        plt.tight_layout()

        out_path = out_dir + f"/parity_{col.lower()}.png"
        plt.savefig(out_path, dpi=120)
        if args.show:
            plt.show()
        else:
            plt.close()

        print(f"Saved {out_path}")


if __name__ == "__main__":
    main()



"""
Create spheres with their specified radii,
then cut them to yield spherical caps
of specified contact angles
"""

from ase.io import write
from os import system
import math
import numpy as np
from ase import Atoms
from ase.build import bulk
import warnings
from ase.data import chemical_symbols
from typing import Union, Optional, List, Tuple
from benchmark.atomistic_utils import (
    scaler,
    create_unit_support,
)

# from functools import lru_cache


# @lru_cache(maxsize=100)
def _calculate_bulk_density(
    element: str,
) -> float:
    """
    Calculate bulk density in atoms/A^3,
    given the element symbol
    """
    atoms = bulk(element)
    volume = atoms.get_volume()
    return len(atoms) / volume


def _atoms_to_radius(
    n_atoms: int,
    element: str,
) -> float:
    """
    Convert number of atoms to radius (of curvature)
    using the formula for a sphere
    returns radius in Ang
    """
    bulk_density = _calculate_bulk_density(
        element=element,
    )
    volume = n_atoms / bulk_density  # A^3
    return ((3 * volume) / (np.pi * 4)) ** (1 / 3)  # A


def _radius_to_atoms(
    radius: float,
    element: str,
) -> int:
    """
    Reverse of atoms_to_radius()
    """
    volume = (4 * np.pi * (radius**3)) / 3  # A^3
    bulk_density = _calculate_bulk_density(
        element=element,
    )  # atoms/A^3
    return volume * bulk_density  # n_atoms


def create_sphere(
    *,  # must use this function by specifying the keywords too, not just values
    element: str,
    n_atoms: Optional[int] = None,
    radius: Optional[float] = None,
) -> Atoms:
    """
    Create a spherical nanoparticle by finding the center of mass of a bulk structure
    then keeping atoms within a specified distance from that center of mass

    Requires:
        n_atoms (int):  required number of atoms in the cluster.
                        will try to fulfil this but may not
        radius (float): required radius of cluster, in Ang.
        element (str):  element type. Mandatory

    Returns:
        atoms_to_keep (Atoms): required spherical nanoparticle/cluster
    """
    element = element.capitalize()
    if radius <= 0.0:
        raise ValueError(f"Invalid radius {radius} supplied")
    if element not in chemical_symbols:
        raise ValueError(f"Invalid element {element} specified")
    if n_atoms is None and radius is None:
        raise ValueError("Provide at least one of radius and n_atoms")
    if n_atoms and radius:
        warnings.warn(
            f"Both radius and n_atoms given; will use only radius", category=UserWarning
        )
        n_atoms = None

    if n_atoms is not None:
        radius = _atoms_to_radius(
            n_atoms=n_atoms,
            element=element,
        )
    diameter = radius * 2
    ##build a bulk cell of safely larger dimensions than the diameter
    base_atoms = bulk(element)
    required_length = 2.5 * diameter
    current_length = min(base_atoms.cell.lengths())
    multiplying_factor = math.ceil(required_length / current_length)
    base_atoms *= (multiplying_factor, multiplying_factor, multiplying_factor)
    ##get the center of mass and draw a circle of the specified diameter around it
    com = base_atoms.get_center_of_mass()
    atoms_to_keep = []
    for index, atom in enumerate(base_atoms):
        position = atom.position
        distance = np.linalg.norm(position - com, ord=2)
        if distance <= radius:
            atoms_to_keep.append(atom)
    atoms_to_keep = Atoms(atoms_to_keep)

    return atoms_to_keep


def cut_particle(atoms: Atoms, angle: float) -> Atoms:
    """
    Cut symetric particle into an asymetric one of a specified contact angle
    """
    atoms = atoms.copy()
    positions = atoms.get_positions()
    z_positions = positions[:, 2]

    ##calculate the min and max position along the z-axis
    z_min = np.min(z_positions)
    z_max = np.max(z_positions)
    z_center = (z_min + z_max) / 2

    angle_radians = np.radians(angle)
    cos_angle = np.cos(
        angle_radians
    )  # calculate the normal vector of the cutting plane

    cut_plane_z = z_center + (z_max - z_center) * cos_angle  # define the cutting plane

    ##select atoms below the cutting plane
    mask = z_positions >= cut_plane_z
    cut_atoms = atoms[mask]

    return cut_atoms



"""
Tests all functions in the by_area script
"""

import pytest
import numpy as np
import warnings
from NanoparticleAtomCounter.by_area import (
    calculate_by_area,
    calculate_areas,
)


def test_calculate_areas():
    """
    No need, here, to test theta since it'll
    be tested when calculate_areas calls calculate_surface_area
    """
    ##test invalid elements
    with pytest.raises(KeyError):
        calculate_areas("Jc", 6.0, 99, None, None)
    ##test negative, low, and zero radii
    with pytest.raises(ValueError):
        calculate_areas("Co", -3.2, 33, None, None)
    with pytest.raises(ValueError):
        calculate_areas("Co", 0.0, 155, (1, 0, 0))
    with pytest.warns(UserWarning):
        calculate_areas("Cu", 1.2, 67, (1, 0, 0))
    ##test valid values
    interface_A, peri_A, surf_A = calculate_areas("Cr", 55, 90, None, (1, 1, 1))
    expected = [8567, 936, 18328]
    for index, region_area in enumerate([interface_A, peri_A, surf_A]):
        assert region_area == pytest.approx(expected[index], abs=2)


def test_calculate_by_area():
    """
    No need, here, to test theta since a downstream function does so;
    Also no need to test element for the same reason
    """
    ##test valid values, using a hemispherical cap,
    # whose equations are simpler to calculate by hand
    peri_atoms, interface_atoms, surf_atoms, total_atoms = calculate_by_area(
        "Ag", 50, 90, (1, 0, 0), (1, 1, 1)
    )
    expected = [106, 833, 2083, 15306]
    for index, region_atoms in enumerate(
        [peri_atoms, interface_atoms, surf_atoms, total_atoms]
    ):
        assert region_atoms == pytest.approx(expected[index], abs=2)



"""
Here, I'll compare the results given by the npatomcounter to those
from making an approximate atomistic model for nanoparticles of an FCC
lattice and flat interface
"""

import subprocess
from ase.io import write
import shlex
import sys
from os import system, environ, makedirs
import numpy as np
import pandas as pd
from itertools import product
from ase import Atoms
from tqdm import tqdm
from typing import List, Union, Tuple, Literal
from ase.visualize import view
import warnings
from argparse import ArgumentParser
from pathlib import Path
from ascii_colors import ASCIIColors
from time import perf_counter
from benchmark.create_spherical_caps import (
    create_sphere,
    cut_particle,
)
from benchmark.atomistic_utils import (
    scaler,
    create_unit_support,
)
import pytest

MIN_ANGLE = 60
MAX_ANGLE = 160
MIN_RADIUS = 10  # Ang
MAX_RADIUS = 30  # Ang
N_ANGLES = 7
N_RADII = 7
PROCESSES = -1
OUTPUT_TRAJECTORY = "atoms.traj"
NP_ELEMENTS = ["Ag"]
SUPPORT_ELEMENTS = ["graphene", "au"]

script_dir = Path(__file__).resolve().parent


def create_outputdir() -> str:
    home_dir = environ["HOME"]
    output_dir = f"{home_dir}/nanoparticle_atom_counter_benchmark/"
    makedirs(output_dir, exist_ok=True)

    return output_dir


def create_trajectory(
    min_angle: float,
    max_angle: float,
    n_angles: int,
    min_radius: float,
    max_radius: float,
    n_radii: int,
    output_trajectory: str,
    np_elements: Union[str, List[str]],
    support_element: Union[str, List[str]],
) -> Tuple[List[float], List[float], List[str], List[str], int]:
    """
    Creates the atomistic model

    Requires:

        min_angle:          lowest contact angle desired
        max_angle:          max contact angle desired
        n_angles:           how many contact angles desired
        min_radius:         lowest radius OF CURVATURE (NOT footprint radius) desired
        max_radius:         figure it out
        n_radii:            figure it out
        output_trajectory:  name of file in which to write the models
        np_elements:        the type of atom of which the nanoparticle is composed,
                            as a list, one for each nanoparticle/support combination
        support_element:   the type of atom of which the support is composed,
                            as a list, one for each nanoparticle/support combination.
                            Should be 'graphene', 'mgo', or 'au'

    Returns:

        contact_angles:     list of contact angles
        radii_angstrom:     list of curvature radii
        nanoparticles:      list of nanoparticle elements
        supports:           list of support elements
        n_calculations:     number of calculations to run
    """

    if max_radius > 40:
        warnings.warn(
            """This will take some time.
                I hope you know what you are doing!""",
            category=UserWarning,
        )

    nanoparticles = [np_elements] if isinstance(np_elements, str) else np_elements
    supports = (
        [support_element] if isinstance(support_element, str) else support_element
    )

    n_calculations = int(n_radii * n_angles * len(nanoparticles) * len(supports))
    requested_system = (
        "Requested system: \n"
        f"All possible combinations of {np_elements} nanoparticles on {support_element}\n"
        f"Nanoparticle curvature radii are from {min_radius} to {max_radius} A,\n"
        f"with contact angles from {min_angle} to {max_angle}"
        f"Number of calculations: {n_calculations}"
    )
    print(requested_system)
    
    radii_angstrom = np.linspace(min_radius, max_radius, n_radii)
    contact_angles = np.linspace(min_angle, max_angle, n_angles)

    atoms_list: List[Atoms] = []

    for r in tqdm(
        radii_angstrom,
        total=len(radii_angstrom),
        desc="creating atomistic models of the requested systems",
    ):
        for theta in contact_angles:
            for nanoparticle in nanoparticles:
                for support in supports:
                    atoms = create_sphere(element=nanoparticle, radius=r)
                    atoms = cut_particle(atoms, theta)
                    unit_support = create_unit_support(support)
                    atoms = scaler(image=atoms, unit_support=unit_support)
                    atoms.info["np_element"] = nanoparticle
                    atoms_list.append(atoms)

    write(output_trajectory, atoms_list)
    print("created and written atoms objects")

    return contact_angles, radii_angstrom, nanoparticles, supports, n_calculations


def run_atomistic(
    processes: int,
    trajectory_file: str,
    radii_angstrom: List[float],
    contact_angles: List[float],
    nanoparticles: List[str],
    supports: List[str],
    input_to_atomcounter: str,
    atomistic_output: str,
    new_atoms_output: str,
    output_dir: str,
) -> None:
    """
    Use the atomistic model to get the number of each kind of atom in each nanoparticle

    Requires:

        processes:              number of parallel processes to use.
        trajectory_file:        the file containing the atomistic models
        radii_angstrom:         list of curvature radii
        contact_angles:         list of contact angles
        nanoparticles:          list of nanoparticle elements
        supports:               list of support elements
        input_to_atomcounter:   name of input file this will generate;
                                will contain the curvature radius, element, and contact angle,
                                which the atomcounter will need
        atomistic_output:       name of output file to be generated, containing the number
                                of each kind of atom in each nanoparticle
        new_atoms_output:       name of trajectory file to be created, in which the perimeter,
                                interface, and surface atoms are colored differently than the bulk
        output_dir:             name of output directory in which all files will be created

    """
    print("discriminating...")
    command = shlex.split(
        f"python {script_dir}/discrimination.py"
        f" -t {trajectory_file} -p {processes} -o {atomistic_output}"
        f" -to {new_atoms_output}"
    )

    discrimination_out = output_dir + "discrimination.out"
    discrimination_err = output_dir + "discrimination.err"

    with open(discrimination_out, "w") as out_f, open(discrimination_err, "w") as err_f:
        subprocess.run(command, stdout=out_f, stderr=err_f, check=True)

    rows = []
    for R, theta, element, _ in product(
        radii_angstrom, contact_angles, nanoparticles, supports
    ):
        rows.append(
            {
                "r (A)": "",
                "R (A)": R,
                "Theta": theta,
                "Element": element.capitalize(),
                "Interface Facet": "(1,0,0)",
                "Surface Facet": "",  # "(1,1,1)",
            }
        )

    df = pd.DataFrame(rows)
    df.to_csv(input_to_atomcounter, index=False)
    print("Finished the atomistic modelling")


def run_atomcounter(
    input_file: str,
    output_file: str,
    output_dir: str
) -> float:
    """
    Runs the atomcounter to get the number of each kind of atom in each nanoparticle

    Requires:

        input_file:     name of input file, having the element, curvature radius, etc
        output_file:    name of output file to be generated
        output_dir:     name of output directory in which all files will be created

    Returns:   

        timing:          timing in seconds
    """

    print("running NanoparticleAtomCounter ...")

    atomcounter_out = output_dir + "atomcounter.out"
    atomcounter_err = output_dir + "atomcounter.err"

    command = shlex.split(f"nanoparticle-atom-count -i {input_file} -o {output_file}")
    with open(atomcounter_out, "w") as out_f, open(atomcounter_err, "w") as err_f:
        subprocess.run(command, stdout=out_f, stderr=err_f, check=True)
        
    timing = None
    with open(atomcounter_out, "r") as t:
        lines = t.readlines()
        last = lines[-1].strip()
        if last.startswith("Calculation took"):
            parts = last.split()
            timing = float(parts[2])
            
    print("Finished running the nanoparticle atom counter")

    return timing #seconds


def plot_parities(atomistic_output: str, atomcounter_output: str, output_dir: str):
    """
    Creates parity plots of the results, comparing the atomistic to the atomcounter

    Requires:
        atomistic_output:       name of output file from the atomistic
        atomcounter_output:     name of output file from the atomcounter
    """
    print("comparing . . .")
    command = shlex.split(
        f"python {script_dir}/plot-parity.py {atomistic_output} "
        f"{atomcounter_output} --output_dir {output_dir}"
    )

    parity_out = output_dir + "parity.out"
    parity_err = output_dir + "parity.err"

    with open(parity_out, "w") as out_f, open(parity_err, "w") as err_f:
        subprocess.run(command, stdout=out_f, stderr=err_f, check=True)


##ikimashou
def main() -> None:
    """
    if any function fails, we exist with a non-zero exit code
    details on the failure will be written in the .err file (by the above functions)
    """
    output_dir = create_outputdir()
    print(f"\n\nWriting all results to {output_dir}\n\n")
    
    traj_file = output_dir + OUTPUT_TRAJECTORY
    input_to_atomcounter = output_dir + "input.csv"
    atomcounter_output = output_dir + "counter.csv"
    atomistic_output = output_dir + "atomistic.csv"
    new_atoms_output = output_dir + "identified.traj"
            
    theory = (
        "- Calculating the total number of atoms by assuming a spherical cap\n"
        "- Calculating perimeter atoms by assuming the interface is an annular ring\n"
        "  (this might introduce some errors)\n"
        "- Calculating surface atoms by assuming the nanoparticle surface is an annulus\n"
        "  (this might also introduce some errors)\n"
     )
        
    ASCIIColors.print(
        theory,
        color=ASCIIColors.color_yellow,
        style=ASCIIColors.style_bold,
        background=ASCIIColors.color_black,
        end="\n\n",
        flush=True,
        file=sys.stdout,
    )
        
    exit_code = 0
    try:
        contact_angles, radii_angstrom, nanoparticles, supports, n_calculations = create_trajectory(
            min_angle=MIN_ANGLE,
            max_angle=MAX_ANGLE,
            n_angles=N_ANGLES,
            min_radius=MIN_RADIUS,
            max_radius=MAX_RADIUS,
            n_radii=N_RADII,
            output_trajectory=traj_file,
            np_elements=NP_ELEMENTS,
            support_element=[SUPPORT_ELEMENTS[0]],
        )
        
        
        run_atomistic(
            processes=PROCESSES,
            trajectory_file=traj_file,
            contact_angles=contact_angles,
            radii_angstrom=radii_angstrom,
            nanoparticles=nanoparticles,
            supports=supports,
            input_to_atomcounter=input_to_atomcounter,
            atomistic_output=atomistic_output,
            new_atoms_output=new_atoms_output,
            output_dir=output_dir,
        )
        
        timing = run_atomcounter(
            input_file=input_to_atomcounter,
            output_file=atomcounter_output,
            output_dir=output_dir,
        )
        
        plot_parities(
            atomistic_output=atomistic_output,
            atomcounter_output=atomcounter_output,
            output_dir=output_dir,
        )
        
        print(f"\n\nAll output written to {output_dir}")
        
        
        explanation = """\
        - input.csv: input file for the NanoparticleAtomCounter
        - counter.csv: atom counts according to the NanoparticleAtomCounter
        - atomistic.csv: atom counts by using an atomistic model
        - atoms.traj: atomistic model
        - identified.traj: atomistic model, with different types of atoms distinguished
        - parity*.png: parity plots comparing the NanoparticleAtomCounter to the atomistic model
        """
        
        speed_benchmark = f"""\
        nanoparticleatomcounter took {timing * 1000} milliseconds to run {n_calculations} calculations
        """
        
        readme = output_dir + "README.md"
        speed = output_dir + "timing.log"
        
        with open(readme, "w") as explain_file, open(speed, "w") as speed_file:
            explain_file.write(explanation)
            speed_file.write(speed_benchmark)
    except Exception as e:
        print(f"benchmarking failed with error ({e}). see .err files in {output_dir} for details")
        exit_code = 1

    return exit_code





"""
Area-based routines for estimating atom counts in supported nanoparticles.

calculates the number of interfacial atoms by looking only at the footprint radius,
which means the contact angle doesn't affect the number of interfacial atoms

Functions
---------
calculate_areas(element: str, r: float, theta: float, facet: tuple) -> tuple
    Calculate areas of the interface, perimeter, and surface
    using:
        1. contact angle (theta)
        2. footprint/interfacial radius (r) in Angstrom
        3. chemical symbol of the nanoparticle's atoms (element)
        4. facet of nanoparticle oriented to support (facet)

calculate_by_area(element: str, r: float, theta: float, facet: tuple) -> tuple
    Main function to do all calculations through the 'area' method
"""

from typing import Tuple, List
import numpy as np
import warnings
from NanoparticleAtomCounter.count_utilities import (
    volume_to_atoms,
    calculate_total_volume,
    area_to_atoms,
    calculate_atomic_density,
    calculate_total_volume,
    calculate_surface_area,
    calculate_constants,
)


def calculate_areas(
    element: str,
    footprint_radius: float,
    theta: float = None,
    interface_facet: Tuple[int, int, int] = None,
    surface_facet: Tuple[int, int, int] = None,
) -> Tuple[float, float, float]:
    """
    Calculate area of:
        1. interface
        2. perimeter
        3. surface
            all in A^2

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        footprint_radius (float):   NP footprint radius. In Ang
        theta (float):              contact angle, in degrees
        interface_facet (Tuple[int,int,int]): facet facing the support
        surface_facet (Tuple[int,int,int]): facet facing vacuum

    Returns:
        interfacial area (float):       interface area. excludes perimeter
        perimeter area (float):         perimeter area.
        surface area (float):           NP outer surface area. excludes interface
    """
    # r: footprint radius (i.e. radius of bottom of the interface),
    # inter_a: interfacial area (using the bottom of the interface).
    # this includes the area of the perimeter
    # peri_a: area of perimeter
    # inner_a: area of interface excluding the perimeter
    # surf_a: area of NP surface, excludes interfacial area
    # d: atomic diameter
    # z: interplanar spacing at the interface (height of the interface)

    _, z, d = calculate_constants(element, facet=interface_facet)

    r = footprint_radius
    footprint_a = np.pi * (r**2)  # A^2
    interface_a = np.pi * ((r - d) ** 2)  # A^2
    peri_a = footprint_a - interface_a  # A^2
    surf_a = calculate_surface_area(element, r, theta, interface_facet=interface_facet)

    return interface_a, peri_a, surf_a  # all in A^2


def calculate_by_area(
    element: str,
    footprint_radius: float,
    theta: float = None,
    interface_facet: Tuple[int, int, int] = None,
    surface_facet: Tuple[int, int, int] = None,
) -> float:
    """
    Main function to do all calculations through the 'area' method

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        footprint_radius (float):   NP footprint radius. In Ang
        theta (float):              contact angle, in degrees
        interface_facet (Tuple[int,int,int]): facet facing the support
        surface_facet (Tuple[int,int,int]): facet facing vacuum

    Returns:
        perimeter_atoms, interfacial_atoms, surface_atoms, all_atoms
    """
    interfacial_area, perimeter_area, NP_surface_area = calculate_areas(
        element,
        footprint_radius,
        theta,
        interface_facet,
        surface_facet,
    )

    perimeter_atoms, interfacial_atoms = [
        area_to_atoms(area=i, element=element, facet=interface_facet)
        for i in [
            perimeter_area,
            interfacial_area,
        ]
    ]

    NP_surface_atoms = area_to_atoms(
        area=NP_surface_area, element=element, facet=surface_facet
    )

    total_volume = calculate_total_volume(footprint_radius, theta)
    total_atoms = volume_to_atoms(total_volume, element, None)

    return perimeter_atoms, interfacial_atoms, NP_surface_atoms, total_atoms



"""
Utilities that the atom counting routines rely on

Functions
---------
calculate_constants:
    Extract molar volume (assumed at equilibrium),
    interplanar spacing, and atomic diameter
calculate_surface_area:
    Calculate area of outer surface of NP
    Excludes the interfacial area and the side area of the perimeter
volume_to_atoms:
    Convert volume (in A^3) to number of atoms
area_to_atoms:
    Convert area (in A^2) to number of atoms
calculate_atomic_density:
    Calculate atomic density (i.e. atoms per unit surface area)
calculate_total_volume:
    Calculate Nanoparticle volume, ASSUMING IT'S A SPHERICAL CAP
alpha:
    a constant needed in the spherical cap equations
beta:
    a constant needed in the spherical cap equations
"""

from typing import Tuple, List
from sys import argv, exit
import numpy as np
import warnings
from functools import lru_cache
from NanoparticleAtomCounter.data import (
    chemical_symbols,
    atomic_numbers,
    atomic_names,
    covalent_radii,
    reference_states,
    molar_volumes,
    interplanar_dist,
    N_A,
)

##extracts molar volume, interplanar spacing, and diameter of the nanoparticle element

# this is the default for interfacial and surface (i.e. gas-oriented) facets
DEFAULT_FACETS = {
    "fcc": "(1, 1, 1)",
    "bcc": "(1, 1, 0)",
    "sc": "(1, 0, 0)",
    "orthorhombic": "(1, 1, 1)",
    "hcp": "(0, 0, 1)",
    "rhombohedral": "(1, 1, 1)",
    "bct": "(1, 1, 1)",
}  # confirmed these to generally be the lowest-energy facets
# (https://next-gen.materialsproject.org/materials and http://crystalium.materialsvirtuallab.org/)


# warnings.filterwarnings("ignore")


@lru_cache(maxsize=10_000)
def calculate_constants(element: str, facet: Tuple[int, int, int] = None) -> float:
    """
    Extract molar volume (assumed at equilibrium),
    interplanar spacing, and atomic diameter

    Requires:
        element (str):      atomic symbol for atom type in the nanoparticle
        facet (Tuple[int,int,int]): facet, whether interfacial or support depends
                                    on the context

    Returns:
        molar volume (float) in A^3/mole,
        interplanar spacing (d) in A
        covalent diameter in A

    Exceptions:
        ValueError if crystal system of the nanoparticle's element is not supported
    """
    ##in case facet is given as (x,y,z) rather than (x, y, z) as expected in data.py
    h, k, l = facet if facet else [None] * 3
    facet = f"({h}, {k}, {l})"
    #    print(facet, element)

    if facet == "(0, 0, 0)":
        raise ValueError(f"Facet cannot be {facet}")

    element = element.capitalize()

    try:
        atomic_number = atomic_numbers[element]
        # if facet not given, pick default
        if h is None:
            reference_state = reference_states[atomic_number]
            crystal_lattice = reference_state["symmetry"]
            facet = DEFAULT_FACETS[crystal_lattice]
            warnings.warn(
                f"Interfacial and-or Surface facet not given, will assume {facet}",
                category=UserWarning,
            )
        covalent_radius = covalent_radii[atomic_number]
        molar_volume = molar_volumes[element]
        interplanar_distance = interplanar_dist[element][facet]
    except KeyError:
        raise KeyError(f"Element type {element} not supported!")
    except Exception as unknown_:
        print(f"Unexpected error occured:\t{unknown_}")

    return molar_volume, interplanar_distance, 2 * covalent_radius


def alpha(theta: int) -> float:
    """
    Constant needed for the spherical cap model equations.
    theta is in degrees
    """
    if theta < 0 or theta > 180:
        warnings.warn(
            f"Invalid value of theta ({theta}) supplied", category=RuntimeWarning
        )

    return 1 / (1 + np.cos(np.radians(theta)))


def beta(theta: int) -> float:
    """
    Constant needed for the spherical cap model equations.
    theta is in degrees
    Will give infinity if theta = 0 or 180
    """
    if theta < 0 or theta > 180:
        raise ValueError(f"Invalid value of theta ({theta}) supplied")
    if theta in [0, 180]:
        raise ValueError(f"Contact angle of {theta} not allowed")

    return (
        (2 + np.cos(np.radians(theta)))
        * (1 - np.cos(np.radians(theta)))
        / (np.sin(np.radians(theta)))
    )


def calculate_surface_area(
    element: str,
    footprint_radius: float,
    theta: float = None,
    interface_facet: Tuple[int, int, int] = None,
) -> float:
    """
    Calculate area of outer surface of NP
    Excludes the interfacial area and the side area of the perimeter

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        interface_facet (Tuple[int,int,int]): facet facing support
        theta (float):              contact angle. degrees

    Returns:
        surface area (float) in A^2
    """
    if theta in [0, 180]:
        raise ValueError(f"Contact angle of {theta} not allowed")
    elif theta > 180 or theta < 0:
        raise ValueError(f"Supplied {theta} is > 180 or < 0. Not allowed")

    r = footprint_radius  # to make things clear
    if r <= 0:
        raise ValueError(f"r ({r}) Ang supplied is <= 0. Invalid")
    if r < 5:
        warnings.warn(
            f"""Small value of r ({r}) Ang supplied; a spherical cap
        approximation may be tenuous""",
            category=UserWarning,
        )

    _, interplanar_spacing, _ = calculate_constants(
        element=element, facet=interface_facet
    )
    z = interplanar_spacing  # to make things clear

    theta_rad = np.radians(theta)
    # how theta changes with shaving off interfacial height
    arg = np.clip(
        np.cos(theta_rad) + ((z / r) * np.sin(theta_rad)), -1.0, 1.0
    )  # this uses not the planar spacing at
    # the surface but that at the interface; and if you derive it, you'll understand why

    #    if arg > 1.0:
    #        warnings.warn(f"""ratio of interplanar spacing and footprint radius too large: ({z/r});
    #        no real cap angle can exist after shaving interface.""",
    #        category = RuntimeWarning)
    #        arg = 1.0
    theta = np.degrees(np.arccos(arg))
    total_area = 2 * np.pi * (r**2) * alpha(theta)

    return total_area  # A^2


def calculate_total_volume(footprint_radius: float, theta: float = None) -> float:
    """
    Calculate Nanoparticle volume, ASSUMING IT'S A SPHERICAL CAP
    Units: A^3

    Requires:
        footprint_radius (float):   NP footprint radius. Ang
        theta (float):              contact angle. degrees

    Returns:
        volume (float):             NP volume. Ang^3
    """
    if theta in [0, 180]:
        raise ValueError(f"Contact angle of {theta} not allowed")
    if footprint_radius <= 5:
        warnings.warn(
            f"""A spherical cap may not work well for this footprint radius({footprint_radius})!
                Mind!""",
            category=UserWarning,
        )

    return (np.pi * (footprint_radius**3)) * alpha(theta) * beta(theta) / 3


def calculate_atomic_density(element: str, facet: Tuple[int, int, int] = None) -> float:
    """
    Calculate atomic density (i.e. atoms per unit surface area)

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        facet (Tuple[int,int,int]): facet, whether interfacial or support depends
                                    on the context

    Returns:
        atomic density (float) in atoms/A^2
    """
    molar_volume, interplanar_spacing, _ = calculate_constants(
        element=element, facet=facet
    )
    atomic_density = interplanar_spacing * N_A / molar_volume

    return atomic_density  # atoms/A^2


def area_to_atoms(area: float, element: str, facet: Tuple[int, int, int] = None) -> int:
    """
    Convert area (in A^2) to number of atoms

    Requires:
        area (float):               Area of region in A^2
        element (str):              atomic symbol for atom type in the nanoparticle
        facet (Tuple[int,int,int]): facet, whether interfacial or support depends
                                    on the context

    Returns:
        n_atoms (int):        Number of atoms in region, rounded to nearest integer
    """
    atomic_density = calculate_atomic_density(element, facet)  # atoms/A^2
    n_atoms = int(np.round(area * atomic_density))

    return n_atoms


def volume_to_atoms(
    volume: float,
    element: str,
    molar_volume: float = None,
) -> int:
    """
    Convert volume (in A^3) to number of atoms

    Requires:
        volume (float):             Volume of enclosed region in A^3
        element (str):              atomic symbol for atom type in the nanoparticle
        molar_volume (float):       molar volume in A^3/mole

    Returns:
        n_atoms (int):              Number of atoms in region, rounded to nearest integer
    """
    if not molar_volume:
        molar_volume, *_ = calculate_constants(element=element)
    #        print(molar_volume)

    bulk_density = N_A / molar_volume  # atom/A^3
    ##set to zero in case interface_volume==0, see by_volume.calculate_volumes
    n_atoms = np.round(volume * bulk_density)
    n_atoms = 0 if np.isnan(n_atoms) else int(n_atoms)

    return n_atoms



"""
benchmark against the atomistic method
"""

from pathlib import Path
import sys
from ascii_colors import ASCIIColors
#import pytest
from benchmark.test_atomcount import main as benchmarking

def main() -> None:
    """
    Run the test_atomcount.py script, which benchmarks
    """
    print("This should take about 5 minutes,\n depending on how many processors you have . . .\n\n")
#    repo_root = Path(__file__).resolve().parents[2]
#    bnchk_dir = repo_root / "benchmark"
#    exit_code = pytest.main(["-s", str(bnchk_dir)])
    exit_code = benchmarking()
    text = "All benchmarking succeeded!" if exit_code == 0 else "Something went wrong! See .err files for details"
    color = ASCIIColors.color_green if exit_code == 0 else ASCIIColors.color_red
    ASCIIColors.print(
        text.upper(),
        color=color,
        style=ASCIIColors.style_bold,
        background=ASCIIColors.color_black,
        end="\n\n",
        flush=True,
        file=sys.stdout,
    )
    sys.exit(exit_code)  # sys.exit(pytest.main(["-s", str(tests_dir)]))


if __name__ == "__main__":
    main()



"""
Main script for estimating atom counts in supported nanoparticles.

Functions
---------
parse_input_data(input_file: str) -> tuple
    Reads the input CSV or XLXS
main:
    run all calculations
"""

import ast
from typing import Tuple, List, Literal
from sys import argv, exit
import numpy as np
import warnings
from NanoparticleAtomCounter.by_volume import calculate_by_volume
from NanoparticleAtomCounter.by_area import calculate_by_area
import pandas as pd
from numpy.typing import NDArray
from os import path
from collections import Counter
import argparse
from time import perf_counter


MODE = "volume"
OUTPUT = "output_atomcounts.csv"


def parse_input_data(
    input_file: str,
) -> Tuple[NDArray, NDArray, NDArray, NDArray, NDArray, NDArray]:
    """
    Reads the input csv or xlxs

    Expected format of the file:
        FIRST ROW ->        r (A),R (A),Theta,Element,Facet
        FOLLOWING ROWS:     a,b,c,d,[x,y,z]

    r:          Footprint (or contact) radius in Angstrom, e.g. 9.1
    R:          Radius of curvature in Angstrom, e.g. 17.12
    Theta:      Contact Angle, e.g. 80
    Element:    Element the Nanoparticle is composed of, e.g. Ag
    Interface Facet:    Facet at the nanoparticle-support interface, given as a tuple, e.g. (1, 1, 1)
    Surface Facet:      Dominant facet at the nanoparticle-gas interface, also as a tuple

    Mandatory:
        1. Theta
        2. r or R. If user supplies both, we choose r and ignore R
        3. Element

    Optional but recommended:
        Interface Facet and Surface Facet
        if not provided, will assume defaults;
        see count_utilities.calculate_constants for the defaults
    """

    ext = path.splitext(input_file)[1].lower()
    if ext in (".csv",):
        df = pd.read_csv(input_file, delimiter=",")
    elif ext in (".xls", ".xlsx"):
        with pd.ExcelFile(input_file) as xls:
            if not xls.sheet_names:
                raise ValueError(
                    f"""
                    '{input_file}' looks like an Excel container but
                    has no worksheets. Is it really an .xlsx file?
                    """
                )
            df = xls.parse(
                sheet_name=xls.sheet_names[0]
            )  # assuming the data is in the first sheet
    #        df = pd.read_excel(input_file, sheet_name = 0) #assuming the data is in the first sheet
    else:
        raise ValueError(
            f"""
        Unrecognized file type: {ext}!
                Requires .csv, .xls, or .xlsx
                """
        )  # first: pip install xlrd openpyxl

    # verify that the cols are labelled properly so we get the correct data
    df.columns = (
        df.columns.str.strip()
    )  # remove leading/trailing spaces. no unifying case or R will become r
    expected = [
        "r (A)",
        "R (A)",
        "Theta",
        "Element",
        "Interface Facet",
        "Surface Facet",
    ]
    # compare as multisets to check both presence *and* counts
    assert Counter(df.columns) == Counter(expected), (
        f"Column mismatch:\n"
        f"  found    = {df.columns.tolist()}\n"
        f"  expected = {expected}"
    )

    rs = df["r (A)"].to_numpy()  # np.nan if no values
    Rs = df["R (A)"].to_numpy()  # np.nan if no values
    thetas = df["Theta"].to_numpy()
    elements = df["Element"].to_numpy()
    interface_facets = (
        df["Interface Facet"]
        .apply(lambda x: ast.literal_eval(x) if pd.notnull(x) else [None] * 3)
        .to_numpy()
    )  # None if no values
    surface_facets = (
        df["Surface Facet"]
        .apply(lambda x: ast.literal_eval(x) if pd.notnull(x) else [None] * 3)
        .to_numpy()
    )  # None if no values

    return rs, Rs, thetas, elements, interface_facets, surface_facets


def main() -> None:
    f"""
    Main function to do all calculations through the selected method

    Requires:
        input_file (str):                   input file name, with full path
        output_file (str):                   output file name, with full path. default = 'output.txt'
        mode (Literal["volume", "area"])    whether to calculate by volume or area. Default = 'volume'


    Returns:
        NONE, but writes out the output file
    """
    parser = argparse.ArgumentParser(
        description="""Given an input file containing:
     (1) NP element type
     (2) footprint radius or radius of curvature
     (3) contact angle,
     (4) (optionally) interface facet
     (5) (optionally) surface facet

    calculates number of surface, perimeter, interfacial, and total atoms"""
    )

    parser.add_argument(
        "--input", "-i", type=str, required=True, help="Path to input file"
    )

    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default=OUTPUT,
        help=f"Path to output file. default = {OUTPUT}",
    )

    parser.add_argument(
        "--mode",
        "-m",
        type=str,
        default=MODE,
        help=f"""Method for calculating: by area ("area") or by volume ("volume").
            default = 'volume' """,
        choices=("volume", "area"),
    )

    start = perf_counter()
    args = parser.parse_args()
    input_file = args.input
    output_file = args.output
    mode = args.mode.lower()

    rs, Rs, thetas, elements, interface_facets, surface_facets = parse_input_data(
        input_file
    )

    if np.all(np.isnan(rs)) or np.all(
        rs == 0
    ):  # if rs is empty or 0, convert Rs into rs
        print("converting r into R...")
        rs = Rs * np.sin(np.radians(thetas))
    elif np.any(np.isnan(rs)):  # if there is any rs value missing, exit
        raise ValueError(f"Some entries for r (A) are missing!")

    # in case some variables have more entries than others
    data = {
        "rs": rs,
        "Rs": Rs,
        "thetas": thetas,
        "elements": elements,
        "interface_facets": interface_facets,
        "surface_facets": surface_facets,
    }

    lengths = {name: len(arr) for name, arr in data.items()}
    unique_lengths = set(lengths.values())

    if len(unique_lengths) > 1:
        warnings.warn(
            f"Inconsistent data lengths: {lengths}. "
            f"Truncating all to the shortest ({min(unique_lengths)})",
            category=UserWarning,
        )

    min_len = min(unique_lengths)  # truncate everything to the same (minimum) length
    for name in data:
        data[name] = data[name][:min_len]

    mode_ = {
        "volume": calculate_by_volume,
        "area": calculate_by_area,
    }

    peri_atoms, inter_atoms, surf_atoms, tot_atoms = zip(
        *[
            mode_[mode](
                element=data["elements"][i],
                footprint_radius=data["rs"][i],
                theta=data["thetas"][i],
                interface_facet=tuple(data["interface_facets"][i]),
                surface_facet=tuple(data["surface_facets"][i]),
                # to be able to cache in calculate_constants(), facets must be immutable (i.e. tuple rather than list)
                # just in case the user had given as [x, y, z] rather than (x, y, z)
                # so, I am converting to a tuple
            )
            for i in range(min_len)
        ]
    )

    print("Success!")

    df = pd.DataFrame(
        {
            "Perimeter": peri_atoms,
            "Interface": inter_atoms,
            "Surface": surf_atoms,
            "Total": tot_atoms,
        }
    )

    df.to_csv(output_file, index=False)
    print(f"Output ({output_file}) written!")
    timing = perf_counter() - start
    print(f"Calculation took {timing} seconds")



if __name__ == "__main__":
    main()



"""
Tests all functions in the by_volume script
"""

import pytest
import numpy as np
import warnings
from NanoparticleAtomCounter.by_volume import (
    calculate_by_volume,
    calculate_volumes,
)


def test_calculate_volumes():
    ##test angles
    for angle in [0, 180, -20, 200]:
        with pytest.raises(ValueError):
            calculate_volumes("Pt", 3.2, angle)
    ##test invalid elements
    with pytest.raises(KeyError):
        calculate_volumes("Fol", 3.2, 33)
    ##test negative, low, and zero radii
    with pytest.raises(ValueError):
        calculate_volumes("Mn", -3.2, 33, (1, 0, 0))
    with pytest.raises(ValueError):
        calculate_volumes("V", 0, 33, (1, 0, 0))
    with pytest.warns(UserWarning):
        calculate_volumes("Ti", 1.2, 134, (1, 0, 0))
    ##test valid values
    interface_V, peri_V, total_V = calculate_volumes("Fe", 66, 90, (1, 1, 1))
    expected = [10442, 888, 602130]
    for index, region_volume in enumerate([interface_V, peri_V, total_V]):
        assert region_volume == pytest.approx(expected[index], abs=2)


def test_calculate_by_volume():
    """
    No need, here, to test theta since a downstream function does so;
    Also no need to test element for the same reason
    """
    ##test valid values, using a hemispherical cap,
    # whose equations are simpler to calculate by hand
    peri_atoms, interface_atoms, surf_atoms, total_atoms = calculate_by_volume(
        "Ni", 150, 90, None, None
    )
    expected = [432, 12740, 25994, 648283]
    for index, region_atoms in enumerate(
        [peri_atoms, interface_atoms, surf_atoms, total_atoms]
    ):
        assert region_atoms == pytest.approx(expected[index], abs=2)



"""
Volume-based routines for estimating atom counts in supported nanoparticles.

Functions
---------
calculate_volumes(element: str, r: float, theta: float, facet: tuple) -> tuple
    Calculate volume of the interfacial region (excluding the perimeter region),
    perimeter region, and the entire nanoparticle
    using:
        1. contact angle (theta)
        2. footprint/interfacial radius (r) in Angstrom
        3. chemical symbol of the nanoparticle's atoms (element)
        4. facet of nanoparticle oriented to support (facet)

calculate_by_volume(element: str, r: float, theta: float, facet: tuple) -> tuple
    Main function to do all calculations through the 'volume' method
"""

from typing import Tuple, List
import numpy as np
import warnings
from NanoparticleAtomCounter.count_utilities import (
    volume_to_atoms,
    area_to_atoms,
    calculate_atomic_density,
    calculate_total_volume,
    calculate_surface_area,
    calculate_constants,
)

# warnings.filterwarnings(action="ignore")


def calculate_volumes(
    element: str,
    footprint_radius: float,
    theta: float = None,
    interface_facet: Tuple[int, int, int] = None,
) -> Tuple[float, float, float]:
    """
    Calculate volume of the interfacial region (excluding the perimeter region),
    perimeter region, and the entire nanoparticle

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        theta (float):              contact angle. degrees
        footprint_radius (float):   NP footprint radius. In Ang
        interface_facet (Tuple[int,int,int]): facet facing the support

    Returns:
        interface_volume, perimeter_volume, total_volume     all in A^3
    """
    # R: radius from spherical cap center,
    # r: footprint radius (i.e. radius of bottom of spherical segment),
    # x: radius of top of spherical segment, z: height of spherical segment,
    # x2: x but excluding the perimeter having thickness == z1,
    # r2: r but excluding the perimeter having thickness == z1,
    # h:
    # if θ>90: distance from sphere center vertically down to top of spherical segment
    # if θ<90: distance from sphere center vertically up to bottom of interface
    # if θ==90: zero
    # z1: atomic diameter
    # z: interplanar spacing at the interface, i.e. this is the height of the interface
    if theta in [0, 180]:
        raise ValueError(f"Contact angle of {theta} not allowed")
    elif theta > 180 or theta < 0:
        raise ValueError(f"Supplied {theta} is > 180 or < 0. Not allowed")

    r = footprint_radius  # to make things clear
    if r <= 0:
        raise ValueError(f"r ({r}) Ang supplied is <= 0. Invalid")
    if r < 5:
        warnings.warn(
            f"""Small value of r ({r}) Ang supplied; a spherical cap
        approximation may be tenuous""",
            category=UserWarning,
        )

    _, z, z1 = calculate_constants(element=element, facet=interface_facet)

    R = r / np.sin(np.radians(theta))
    # r2 = radial spacing
    r2 = np.clip(
        r - z1, 0, None
    )  # no reliable formula for radial spacing. assumed to be = atomic diameter
    # clip so it doesn't become negative for extremely small footprint radii
    ##Find h and x
    if theta > 90:
        h = np.sqrt(R**2 - r**2) - z  # thus spake pythagoras
        x = np.sqrt(R**2 - h**2)  # eqn 1
        dome_height = h + z + R
        x2 = np.sqrt(
            (2 * h * z) + (z**2) + (r2**2)
        )  # from eqn1, noting that R^2 = (h+z)^2 + r^2 and then substituting r2 for r
    else:
        h = np.clip(np.sqrt(R**2 - r**2), 0, None)  # clip so h==0 if theta == 90
        dome_height = R + h
        x = np.sqrt(R**2 - ((h + z) ** 2))  # eqn 2
        x2 = np.sqrt(
            (h**2) - ((h + z) ** 2) + (r2**2)
        )  # can have invalid values; see below notes


    # Formula from https://en.wikipedia.org/wiki/Spherical_segment
    segment_volume = np.pi * z * ((3 * (r**2 + x**2)) + z**2) / 6
    interface_volume = np.pi * z * ((3 * (r2**2 + x2**2)) + z**2) / 6
    total_volume = calculate_total_volume(r, theta)

    # if x2 was negative, then it means for the given combination of r and theta, if you shave off z from r
    # then too much will be shaved off from the top of the spherical segment that its radius (x2) becomes negative
    # this implies that we can't have any non-perimeter atoms for such a system
    # for a given r, will be more likely for smaller theta
    # for a given theta, will be more likely for smaller r
    if np.isnan(interface_volume):
        warnings.warn(
            f"""footprint radius ({r}) and-or θ ({theta}) too small.
        There is no non-perimeter for the given combination of r and θ""",
            category=RuntimeWarning,
        )
        interface_volume = 0

    perimeter_volume = segment_volume - interface_volume

    return interface_volume, perimeter_volume, total_volume


def calculate_by_volume(
    element: str,
    footprint_radius: float,
    theta: float = None,
    interface_facet: Tuple[int, int, int] = None,
    surface_facet: Tuple[int, int, int] = None,
) -> Tuple[int, int, int, int]:
    """
    Main function to do all calculations through the 'volume' method

    Units: A^3

    Requires:
        element (str):              atomic symbol for atom type in the nanoparticle
        theta (float):              contact angle. degrees
        footprint_radius (float):   NP footprint radius. In Ang
        interface_facet (Tuple[int,int,int]): facet facing the support
        surface_facet (Tuple[int,int,int]): facet facing vacuum

    Returns:
        perimeter_atoms, interfacial_atoms, surface_atoms, all_atoms
    """

    interfacial_volume, perimeter_volume, total_volume = calculate_volumes(
        element=element,
        footprint_radius=footprint_radius,
        theta=theta,
        interface_facet=interface_facet,
    )

    interfacial_atoms, perimeter_atoms, total_atoms = [
        volume_to_atoms(
            volume=i,
            element=element,
            molar_volume=None,
        )
        for i in [interfacial_volume, perimeter_volume, total_volume]
    ]

    surface_area = calculate_surface_area(
        element=element,
        footprint_radius=footprint_radius,
        theta=theta,
        interface_facet=interface_facet,
    )

    surface_atoms = area_to_atoms(
        area=surface_area, element=element, facet=surface_facet
    )

    return perimeter_atoms, interfacial_atoms, surface_atoms, total_atoms



##generated this setup.py using setup-py-cli
##from https://pypi.org/project/setup-py-cli/

from setuptools import setup, find_packages
import io
import os
from NanoparticleAtomCounter import __version__

here = os.path.abspath(os.path.dirname(__file__))

# read the contents of your README.md
with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="NanoparticleAtomCounter",
    version=__version__,
    author="Gbolagade Olajide, Tibor Szilvasi",
    author_email="giolajide@crimson.ua.edu, tszilvasi@crimson.ua.edu",
    description="Estimates atom counts in monometallic nanoparticles"
    " given radius and contact angle",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/szilvasi-group/NanoparticleAtomCounter#readme",
    packages=find_packages(where="."),
    # specify your dependencies here
    python_requires=">=3.10,<3.15",
    install_requires=[
        "numpy>=1.20,<2.3",
        "pandas>=1.4",
        "streamlit>=1.4,<2",
        "openpyxl>=3.1",
        "xlrd>=2.0",
    ],
    extras_require={
        
        "test": [
            "pytest>=8,<9",
            "ascii-colors",
        ],
        
        "benchmark": [
            "joblib>=1.5,<2",
            "ase>=3.22,<3.27",
            "ascii-colors",
            "tqdm>=4.66,<5",
            "pytest>=8,<9",
        ],

        "dev": [
            "joblib>=1.5,<2",
            "ase>=3.22,<3.27",
            "ascii-colors",
            "tqdm>=4.66,<5",
            "pytest>=8,<9",
        ],
        
    },
    # enable the CLI: `nanoparticle-atom-count`
    entry_points={
        "console_scripts": [
            "nanoparticle-atom-count="
            "NanoparticleAtomCounter.cli.atom_count:main",  # main script
            
            "atom-count-test="
            "NanoparticleAtomCounter.cli.run_tests:main",  # tests script
            
            "atom-count-benchmark="
            "NanoparticleAtomCounter.cli.benchmark:main", # benchmark script
        ],
    },
    classifiers=[
        ##https://pypi.org/classifiers/
        # Project maturity
        "Development Status :: 4 - Beta",
        # Audience & topic
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        # License
        "License :: OSI Approved :: MIT License",
        # Supported Python versions
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords=["nanoparticle", "tem", "electron microscopy", "active site", "catalyst"],
    license="MIT",
    include_package_data=True,
    project_urls={
        "Bug Tracker": "https://github.com/szilvasi-group/NanoparticleAtomCounter/issues",
        "Documentation": "https://github.com/szilvasi-group/NanoparticleAtomCounter#readme",
    },
)



"""
Tests all functions in the count_utilities script
"""

import pytest
import numpy as np
import warnings
from NanoparticleAtomCounter.count_utilities import (
    volume_to_atoms,
    beta,
    alpha,
    area_to_atoms,
    calculate_atomic_density,
    calculate_total_volume,
    calculate_surface_area,
    calculate_constants,
    DEFAULT_FACETS,
)

TOLERANCE = 1e-4


def test_default_facets():
    assert DEFAULT_FACETS["hcp"] == "(0, 0, 1)"


def test_calculate_constants():
    pd_molarvol, pd100_planarspacing, pd_dia = calculate_constants("Pd", (1, 0, 0))
    assert pd_molarvol == pytest.approx(8.86216261990501e24, abs=TOLERANCE)
    assert pd100_planarspacing == pytest.approx(1.945, abs=TOLERANCE)
    assert pd_dia == pytest.approx(2.78, abs=TOLERANCE)


def test_alpha():
    assert alpha(180) == np.inf
    assert alpha(90) == pytest.approx(1.0, abs=TOLERANCE)


def test_beta():
    for angle in [0, 180]:
        with pytest.raises(ValueError):
            beta(angle)
    assert beta(90) == pytest.approx(2.0, abs=TOLERANCE)


def test_calculate_surface_area():
    ##test angles
    for angle in [0, 180, -20, 200]:
        with pytest.raises(ValueError):
            calculate_surface_area("Pt", 3.2, angle, (1, 0, 0))
    ##test invalid elements
    with pytest.raises(KeyError):
        calculate_surface_area("Px", 3.2, 33, (1, 0, 0))
    ##test negative, low, and zero radii
    with pytest.raises(ValueError):
        calculate_surface_area("Pd", -3.2, 33, (1, 0, 0))
    with pytest.raises(ValueError):
        calculate_surface_area("Pd", 0, 33, (1, 0, 0))
    with pytest.warns(UserWarning):
        calculate_surface_area("Pd", 1.2, 134, (1, 0, 0))
    ##test valid values
    area = calculate_surface_area("Pd", 70, 122, None)
    assert area > 0


def test_calculate_total_volume():
    vol = calculate_total_volume(9, 99)
    assert vol > 0


def test_calculate_atomic_density():
    assert calculate_atomic_density("Ag", (1, 1, 1)) == pytest.approx(
        0.1380551931635095, abs=TOLERANCE
    )


def test_area_to_atoms():
    assert area_to_atoms(200, "Au", (1, 1, 1)) == pytest.approx(27, abs=1)


def test_volume_to_atoms():
    assert volume_to_atoms(1000, "Cr", None) == pytest.approx(83, abs=1)
